@@ -1,8 +1,18 @@
 print("Interacting with reality...")
 import numpy as np
+try:
+    import cupy as cp
+    print("GPU accelerated computing enabled. You can set Render('cupy') to benefit GPU utils.")
+    cupy_enabled=True
+except ImportError:
+    cupy_enabled=False
+    cp=np
+    print("Running in CPU mode, set Render('numpy') to use it.")
+
+
 import sys
 import glimy.geo
-
+import psutil
 import warnings
 import time
 import matplotlib.pyplot as plt
@@ -15,6 +25,9 @@ Z_0=376.730313
 e_0=8.8541878128e-12
 mu_0=1.25663706212e-6
 
+e_0_2=e_0*2
+mu_0_2=2*mu_0
+
 class Continuum(object):
     def __init__(self, grid_size, ds):
         
@@ -60,6 +73,8 @@ class Continuum(object):
         self.__bound_bool=False
         self.__rendered=False
         
+        self.__built_for=None
+        
 
         
         
@@ -122,7 +137,9 @@ class Continuum(object):
                 __+=m/np.linalg.norm(l-x*self.__ds)
             return __
 
-        self.__Jdt=1/np.sqrt(1+g_c_2*np.fromfunction(np.vectorize(gp), self.__grid_size, dtype=float))
+        self.__Jdt=1/(1+g_c_2*np.fromfunction(np.vectorize(gp), self.__grid_size, dtype=float))
+        if self.__cupy_enabled:
+            self.__Jdt_cp=cp.array(self.__Jdt)
         
 
     def __emo_video_creator(self):
@@ -131,6 +148,7 @@ class Continuum(object):
             shape=self.__grid_size+(self.__dim,self.__dim)
             self.__e=np.full(shape,np.eye(self.__dim))
             self.__mu=np.full(shape,np.eye(self.__dim))
+
         else:
             shape=self.__grid_size
             self.__e=np.ones(shape)
@@ -141,6 +159,13 @@ class Continuum(object):
         
         if self.__conductivity_m:
             self.__sigma_m=np.zeros(self.__grid_size)
+            
+            
+        if self.__cupy_enabled:
+            self.__e_cp=cp.array(self.__e)
+            self.__mu_cp=cp.array(self.__mu)
+            if self.__conductivity:self.__sigma_cp=cp.array(self.__sigma)
+            if self.__conductivity_m:self.__sigma_m_cp=cp.array(self.__sigma_m)
         
         
         
@@ -152,9 +177,11 @@ class Continuum(object):
                      e[i,i]=element.fielder[0]
                      mu[i,i]=element.fielder[1]
                 
-                fielder=e,mu,element.fielder[2]
+                fielder=e,mu,element.fielder[2]#Needs Future Action(0): sigma_m fielders are not actually rendered by renderer. If necessary physics both enabled for anisotropic materials, the action should be modified
+
             else:
                 fielder=element.fielder
+
             mfs=[a[1]-a[0] for a in element.coverage]
             if not self.__anisotropy:
                 modified_array=np.full(mfs,False,dtype=bool)
@@ -167,6 +194,7 @@ class Continuum(object):
                 ind=index+sl
                 if element.isIn(ind):                    
                     modified_array[index]=ad
+
             
 
             action=(element.layer, element.coverage, modified_array, fielder)
@@ -214,24 +242,32 @@ class Continuum(object):
                     self.__mu[indices]*=np.logical_not(action[2])
                     self.__mu[indices]+=np.full(mfs+[self.__dim, self.__dim],action[3][1])*action[2]
                     
+
                     if self.__conductivity:
                         
                         self.__sigma[indices]*=np.logical_not(action[2][:,:,:,0,0])
                         tg_a=np.full(mfs,action[3][2])*action[2][:,:,:,0,0]
                         self.__sigma[indices]+=tg_a
-                        
+
                     if self.__conductivity_m:
                             
                         self.__sigma_m[indices]*=np.logical_not(action[2][:,:,:,0,0])
                         tg_a=np.full(mfs,action[3][3])*action[2][:,:,:,0,0]
                         self.__sigma_m[indices]+=tg_a
 
-                    
+      
+                    if self.__cupy_enabled:
+                        self.__e_cp[indices]=cp.array(self.__e[indices])
+                        self.__mu_cp[indices]=cp.array(self.__mu[indices])
+                        if self.__conductivity:self.__sigma_cp[indices]=cp.array(self.__sigma[indices])
+                        if self.__conductivity_m:self.__sigma_m_cp[indices]=cp.array(self.__sigma_m[indices])
+
                 else:
                     self.__e[indices]*=np.logical_not(action[2])
                     tg_a=np.full(mfs,action[3][0])*action[2]
                     self.__e[indices]+=tg_a
                     
+                    
                     self.__mu[indices]*=np.logical_not(action[2])
                     tg_a=np.full(mfs,action[3][1])*action[2]
                     self.__mu[indices]+=tg_a
@@ -240,21 +276,42 @@ class Continuum(object):
                         self.__sigma[indices]*=np.logical_not(action[2])
                         tg_a=np.full(mfs,action[3][2])*action[2]
                         self.__sigma[indices]+=tg_a
-                        
+                                                
                     if self.__conductivity_m:    
                         self.__sigma_m[indices]*=np.logical_not(action[2])
                         tg_a=np.full(mfs,action[3][3])*action[2]
                         self.__sigma_m[indices]+=tg_a
+                                                
+                        
+                    if self.__cupy_enabled:
+                        self.__e_cp[indices]=cp.array(self.__e[indices])
+                        self.__mu_cp[indices]=cp.array(self.__mu[indices])
+                        if self.__conductivity:self.__sigma_cp[indices]=cp.array(self.__sigma[indices])
+                        if self.__conductivity_m:self.__sigma_m_cp[indices]=cp.array(self.__sigma_m[indices])
                     
         if self.__conductivity:
-            self.__sigma_ex=self.__sigma*self.__Jdt*self.__dt/2/e_0/self.__e
+            if self.__cupy_enabled:
+                self.__sigma_ex=self.__sigma_cp*self.__Jdt_cp*self.__dt/e_0_2/self.__e_cp
+            else:
+                self.__sigma_ex=self.__sigma*self.__Jdt*self.__dt/e_0_2/self.__e
         if self.__conductivity_m:
-            self.__sigma_mux=self.__sigma_m*self.__Jdt*self.__dt/2/mu_0/self.__mu
+            if self.__cupy_enabled:
+                self.__sigma_mux=self.__sigma_m_cp*self.__Jdt_cp*self.__dt/mu_0_2/self.__mu_cp
+            else:
+                self.__sigma_mux=self.__sigma_m*self.__Jdt*self.__dt/mu_0_2/self.__mu
             
         
-        if self.__anisotropy:            
-            self.__e_inv=np.linalg.inv(self.__e)
-            self.__mu_inv=np.linalg.inv(self.__mu)
+        if self.__anisotropy:
+            if self.__cupy_enabled:
+                self.__e_inv=cp.linalg.inv(self.__e_cp)
+                self.__mu_inv=cp.linalg.inv(self.__mu_cp)
+            else:
+                self.__e_inv=np.linalg.inv(self.__e)
+                self.__mu_inv=np.linalg.inv(self.__mu)
+
+        if self.__cupy_enabled:
+            cp._default_memory_pool.free_all_blocks()
+
             
             
                     
@@ -264,51 +321,6 @@ class Continuum(object):
         
         change=False
         
-        if self.__instance in self.__start_frames:
-            change=True
-            for action in self.__video_starters[self.__instance]:
-                mfs=[a[1]-a[0] for a in action[1]]
-                indices=tuple([slice(k[0], k[1]) for k in action[1]])
-
-                if self.__anisotropy:
-
-                    self.__e[indices]*=np.logical_not(action[2])
-                    self.__e[indices]+=np.full(mfs+[self.__dim, self.__dim],action[3][0])*action[2]
-                    
-                    self.__mu[indices]*=np.logical_not(action[2])
-                    self.__mu[indices]+=np.full(mfs+[self.__dim, self.__dim],action[3][1])*action[2]
-                    
-                    if self.__conductivity:
-                        self.__sigma[indices]*=np.logical_not(action[2][:,:,:,0,0])
-                        tg_a=np.full(mfs,action[3][2])*action[2][:,:,:,0,0]
-                        self.__sigma[indices]+=tg_a
-                        
-                    if self.__conductivity_m:
-                        self.__sigma_m[indices]*=np.logical_not(action[2][:,:,:,0,0])
-                        tg_a=np.full(mfs,action[3][3])*action[2][:,:,:,0,0]
-                        self.__sigma_m[indices]+=tg_a
-                    
-                else:
-                    
-                    self.__e[indices]*=np.logical_not(action[2])
-                    tg_a=np.full(mfs,action[3][0])*action[2]
-                    self.__e[indices]+=tg_a
-                    
-                    self.__mu[indices]*=np.logical_not(action[2])
-                    tg_a=np.full(mfs,action[3][1])*action[2]
-                    self.__mu[indices]+=tg_a
-                    
-                    if self.__conductivity:    
-                        self.__sigma[indices]*=np.logical_not(action[2])
-                        tg_a=np.full(mfs,action[3][2])*action[2]
-                        self.__sigma[indices]+=tg_a
-                        
-                    if self.__conductivity_m:    
-                        self.__sigma_m[indices]*=np.logical_not(action[2])
-                        tg_a=np.full(mfs,action[3][3])*action[2]
-                        self.__sigma_m[indices]+=tg_a
-                    
-                    
         if self.__instance in self.__end_frames:
             change=True
             for action in self.__video_terminators[self.__instance]:
@@ -351,18 +363,90 @@ class Continuum(object):
                         self.__sigma_m[indices]*=np.logical_not(action[2])
                         tg_a=np.full(mfs,0)*action[2]
                         self.__sigma_m[indices]+=tg_a
+
+                if self.__cupy_enabled:
+                    self.__e_cp[indices]=cp.array(self.__e[indices])
+                    self.__mu_cp[indices]=cp.array(self.__mu[indices])
+                    if self.__conductivity:self.__sigma_cp[indices]=cp.array(self.__sigma[indices])
+                    if self.__conductivity_m:self.__sigma_m_cp[indices]=cp.array(self.__sigma_m[indices])
+
+        
+        if self.__instance in self.__start_frames:
+            change=True
+            for action in self.__video_starters[self.__instance]:
+                mfs=[a[1]-a[0] for a in action[1]]
+                indices=tuple([slice(k[0], k[1]) for k in action[1]])
+
+                if self.__anisotropy:
+
+                    self.__e[indices]*=np.logical_not(action[2])
+                    self.__e[indices]+=np.full(mfs+[self.__dim, self.__dim],action[3][0])*action[2]
+                    
+                    self.__mu[indices]*=np.logical_not(action[2])
+                    self.__mu[indices]+=np.full(mfs+[self.__dim, self.__dim],action[3][1])*action[2]
+                    
+                    if self.__conductivity:
+                        self.__sigma[indices]*=np.logical_not(action[2][:,:,:,0,0])
+                        tg_a=np.full(mfs,action[3][2])*action[2][:,:,:,0,0]
+                        self.__sigma[indices]+=tg_a
+                        
+                    if self.__conductivity_m:
+                        self.__sigma_m[indices]*=np.logical_not(action[2][:,:,:,0,0])
+                        tg_a=np.full(mfs,action[3][3])*action[2][:,:,:,0,0]
+                        self.__sigma_m[indices]+=tg_a
+
+                else:
+                    
+                    self.__e[indices]*=np.logical_not(action[2])
+                    tg_a=np.full(mfs,action[3][0])*action[2]
+                    self.__e[indices]+=tg_a
+                    
+                    self.__mu[indices]*=np.logical_not(action[2])
+                    tg_a=np.full(mfs,action[3][1])*action[2]
+                    self.__mu[indices]+=tg_a
+                    
+                    if self.__conductivity:    
+                        self.__sigma[indices]*=np.logical_not(action[2])
+                        tg_a=np.full(mfs,action[3][2])*action[2]
+                        self.__sigma[indices]+=tg_a
+                        
+                    if self.__conductivity_m:    
+                        self.__sigma_m[indices]*=np.logical_not(action[2])
+                        tg_a=np.full(mfs,action[3][3])*action[2]
+                        self.__sigma_m[indices]+=tg_a
+                        
+                if self.__cupy_enabled:
+                    self.__e_cp[indices]=cp.array(self.__e[indices])
+                    self.__mu_cp[indices]=cp.array(self.__mu[indices])
+                    if self.__conductivity:self.__sigma_cp[indices]=cp.array(self.__sigma[indices])
+                    if self.__conductivity_m:self.__sigma_m_cp[indices]=cp.array(self.__sigma_m[indices])
+
         
                 
         if change:
             if self.__conductivity:
-                self.__sigma_ex=self.__sigma*self.__Jdt*self.__dt/2/e_0/self.__e
+                if self.__cupy_enabled:
+                    self.__sigma_ex=self.__sigma_cp*self.__Jdt*self.__dt/e_0_2/self.__e
+                else:
+                    self.__sigma_ex=self.__sigma*self.__Jdt*self.__dt/e_0_2/self.__e
             if self.__conductivity_m:
-                self.__sigma_mux=self.__sigma_m*self.__Jdt*self.__dt/2/mu_0/self.__mu
-            
+                if self.__cupy_enabled:
+                    self.__sigma_mux=self.__sigma_m_cp*self.__Jdt*self.__dt/mu_0_2/self.__mu
+                else:
+                    self.__sigma_mux=self.__sigma_m*self.__Jdt*self.__dt/mu_0_2/self.__mu
+                
             
             if self.__anisotropy:
-                self.__e_inv=np.linalg.inv(self.__e)
-                self.__mu_inv=np.linalg.inv(self.__mu)
+                if self.__cupy_enabled:
+                    self.__e_inv=cp.linalg.inv(self.__e_cp)
+                    self.__mu_inv=cp.linalg.inv(self.__mu_cp)
+                else:
+                    self.__e_inv=np.linalg.inv(self.__e)
+                    self.__mu_inv=np.linalg.inv(self.__mu)
+                    
+        if self.__cupy_enabled:
+            cp._default_memory_pool.free_all_blocks()
+
             
         
 
@@ -372,7 +456,18 @@ class Continuum(object):
     def export_field(self):
         return self.__E, self.__H
 
-    def build(self,verbose=1):
+    def build(self,verbose=1,gpu_flag=False):
+        if (not cupy_enabled and gpu_flag):
+            raise NotImplementedError("GPU resources are not available. Set gpu_flag=False")
+        
+        self.__cupy_enabled=bool(gpu_flag)
+        
+        if gpu_flag==True:
+            try:device=cp.cuda.Device(0);print(f"Building for GPU{device} Render stream.\n{int(device.mem_info[0]/1048576)}MiB({device.mem_info[0]/device.mem_info[1]*100:.1f}%) VRAM available");self.__built_for="cupy";
+            except:pass
+        else:mem=psutil.virtual_memory();print(f"Building for CPU Render stream.\n{int(mem.available/1048576)}MiB({100-mem.percent}%) RAM available");self.__built_for="numpy";
+        if cupy_enabled and not gpu_flag:warnings.warn("GPU is available. You can use GPU for accelerating the computations significantly faster. Call build(gpu_flag=1) for utilizing GPU resources", UserWarning)
+        
         if verbose:
             start=time.time()
         self.__bound=None
@@ -392,6 +487,10 @@ class Continuum(object):
         self.__E=np.zeros((3,)+self.__grid_size)
         self.__H=np.zeros((3,)+self.__grid_size)
         
+        if gpu_flag==True:
+            self.__E_cp=cp.zeros((3,)+self.__grid_size)
+            self.__H_cp=cp.zeros((3,)+self.__grid_size)
+        
         if verbose:
             rt=time.time()-start
 
@@ -971,7 +1070,7 @@ class Continuum(object):
                 
                 if self.__conductivity_m:
                     
-                    self.__H=self.__H *(1-self.__sigma_mux)/(1+self.__sigma_mux)+ pre_E()*Z_c2/self.__e/(1+self.__sigma_mux)
+                    self.__H=self.__H *(1-self.__sigma_mux)/(1+self.__sigma_mux)+ pre_E()*Z_c2/self.__mu/(1+self.__sigma_mux)
                 else:
                     self.__H+=pre_E()*Z_c2/self.__mu
                     
@@ -1120,6 +1219,196 @@ class Continuum(object):
 
 
                             
+        if obs:
+            return obs_array
+        
+    def __cupy_renderer(self, *args):
+        if args[1]:
+            obs=True
+            n_obs=len(args[1])
+            obs_array=np.empty((args[0],n_obs,3))
+            
+        else:
+            obs=False
+            
+        if self.__dim==2:
+            Z_c1=Z_0/2**.5*self.__Jdt_cp
+            Z_c2=1/Z_0/2**.5*self.__Jdt_cp
+            arr_shape=self.__E.shape
+
+
+            def pre_E():
+                ind = cp.zeros(arr_shape)
+            
+                ind[0,:,:-1] += self.__E_cp[2,:,1:] - self.__E_cp[2,:,:-1]
+                ind[1,:-1,:] -= self.__E_cp[2,1:,:] - self.__E_cp[2,:-1,:]
+            
+                return ind
+            
+            
+            def pre_H():
+                ind = cp.zeros(arr_shape)
+                
+                ind[2,:,1:] += self.__H_cp[0,:,1:] - self.__H_cp[0,:,:-1]
+                ind[2,1:,:] -= self.__H_cp[1,1:,:] - self.__H_cp[1,:-1,:]
+                return ind
+            
+            start_time=time.time()
+            
+            
+            for t in range(args[0]):
+                
+                
+                if self.__conductivity_m:
+                    
+                    self.__H_cp=self.__H_cp *(1-self.__sigma_mux)/(1+self.__sigma_mux)+ pre_E()*Z_c2/self.__mu_cp/(1+self.__sigma_mux)
+                else:
+                    self.__H_cp+=pre_E()*Z_c2/self.__mu_cp
+                                        
+                    
+                    
+                if self.__conductivity:
+                    self.__E_cp=self.__E_cp*(1-self.__sigma_ex)/(1+self.__sigma_ex)+pre_H()*Z_c1/self.__e_cp/(1+self.__sigma_ex)
+                    
+                else:
+                    self.__E_cp+=pre_H()*Z_c1/self.__e_cp
+
+
+                for energizer in self.__energizers:
+                    self.__E_cp[:,energizer.location[0],energizer.location[1]]=cp.array(energizer(t))
+                
+                self.__update_grid()
+
+                
+                if t%10==0:
+                    self.__create_progress_bar_with_ETA(start_time,t,args[0])
+                    
+                if obs:
+                    for i in range(n_obs):
+                        for ind in range(3):
+                            obs_array[t,i,ind]=self.__E_cp[ind][args[1][i]]
+                            
+                if self.__bound_bool:
+                    if self.__bound[2]==1:
+                        self.__E_cp[2,:,0]=-self.__E_cp[2,:,1]
+                        self.__H_cp[0,:,0]=-self.__H_cp[0,:,1]
+                    
+                    if self.__bound[3]==1:
+                        self.__E_cp[2,:,-1]=-self.__E_cp[2,:,-2]
+                        self.__H_cp[0,:,-2]=-self.__H_cp[0,:,-3]
+                        
+                    if self.__bound[1]==1:
+                        self.__E_cp[2,-1]=-self.__E_cp[2,-2]
+                        self.__H_cp[1,-2]=-self.__H_cp[1,-3]
+                        
+                    if self.__bound[0]==1:
+                        self.__E_cp[2,0]=-self.__E_cp[2,1]
+                        self.__H_cp[1,0]=-self.__H_cp[1,1]
+
+
+
+        elif self.__dim==3:
+            
+            
+            Z_c1=Z_0/3**.5
+            Z_c2=1/Z_0/3**.5
+            arr_shape=self.__E.shape
+            
+            def pre_E():
+                ind = cp.zeros(arr_shape)
+            
+                ind[0,:,:-1,:] += self.__E_cp[2,:,1:,:] - self.__E_cp[2,:,:-1,:]
+                ind[0,:,:,:-1] -= self.__E_cp[1,:,:,1:] - self.__E_cp[1,:,:,:-1]
+                ind[1,:,:,:-1] += self.__E_cp[0,:,:,1:] - self.__E_cp[0,:,:,:-1]
+                ind[1,:-1,:,:] -= self.__E_cp[2,1:,:,:] - self.__E_cp[2,:-1,:,:]
+                ind[2,:-1,:,:] += self.__E_cp[1,1:,:,:] - self.__E_cp[1,:-1,:,:]
+                ind[2,:,:-1,:] -= self.__E_cp[0,:,1:,:] - self.__E_cp[0,:,:-1,:]
+            
+                return ind
+            
+            
+            def pre_H():
+                ind = cp.zeros(arr_shape)
+                
+                ind[0,:,:,1:] += self.__H_cp[1,:,:,1:] - self.__H_cp[1,:,:,:-1]
+                ind[0,:,1:,:] -= self.__H_cp[2,:,1:,:] - self.__H_cp[2,:,:-1,:]
+                ind[1,1:,:,:] += self.__H_cp[2,1:,:,:] - self.__H_cp[2,:-1,:,:]
+                ind[1,:,:,1:] -= self.__H_cp[0,:,:,1:] - self.__H_cp[0,:,:,:-1]
+                ind[2,:,1:,:] += self.__H_cp[0,:,1:,:] - self.__H_cp[0,:,:-1,:]
+                ind[2,1:,:,:] -= self.__H_cp[1,1:,:,:] - self.__H_cp[1,:-1,:,:]        
+                return ind
+            
+            start_time=time.time()
+            
+            coord_E=self.__Jdt_cp*Z_c1
+            coord_H=self.__Jdt_cp*Z_c2
+                
+            
+            
+            for t in range(args[0]):
+                    
+                if self.__anisotropy:
+                    self.__H_cp+=cp.einsum("klmij,jklm->iklm",self.__mu_inv,pre_E())*coord_H
+                    self.__E_cp+=cp.einsum("klmij,jklm->iklm",self.__e_inv,pre_H())*coord_E
+                        
+                else:
+                    if self.__conductivity_m:
+                        self.__H_cp=self.__H_cp*(1-self.__sigma_mux)/(1+self.__sigma_mux)+pre_E()*coord_H/self.__mu_cp/(1+self.__sigma_mux)
+
+                    else:
+                        self.__H_cp+=pre_E()*coord_H/self.__mu_cp
+                        
+                        
+                    if self.__conductivity:
+                        self.__E_cp=self.__E_cp*(1-self.__sigma_ex)/(1+self.__sigma_ex)+pre_H()*coord_E/self.__e_cp/(1+self.__sigma_ex)
+                    else:
+                        self.__E_cp+=pre_H()*coord_E/self.__e_cp
+                
+                
+                for energizer in self.__energizers:
+                    self.__E_cp[:,energizer.location[0],energizer.location[1],energizer.location[2]]=cp.array(energizer(t))
+                
+                self.__update_grid()
+                
+                self.__create_progress_bar_with_ETA(start_time,t,args[0])
+                    
+                if obs:
+                    for i in range(n_obs):
+                        for ind in range(3):
+                            obs_array[t,i,ind]=self.__E_cp[ind][args[1][i]]
+                            
+        
+                if self.__bound_bool:
+                    # Boundaries in the x-direction
+                    if self.__bound[0] == 1:
+                        self.__E_cp[:, 0, :, :] = self.__E_cp[:, 1, :, :]
+                        self.__H_cp[:, 0, :, :] = self.__H_cp[:, 1, :, :]
+                
+                    if self.__bound[1] == 1:
+                        self.__E_cp[:, -2, :, :] = -self.__E_cp[:, -3, :, :]
+                        self.__H_cp[:, -2, :, :] = -self.__H_cp[:, -3, :, :]
+                
+                    # Boundaries in the y-direction
+                    if self.__bound[2] == 1:
+                        self.__E_cp[:, :, 0, :] = self.__E_cp[:, :, 1, :]
+                        self.__H_cp[:, :, 0, :] = self.__H_cp[:, :, 1, :]
+                
+                    if self.__bound[3] == 1:
+                        self.__E_cp[:, :, -2, :] = -self.__E_cp[:, :, -3, :]
+                        self.__H_cp[:, :, -2, :] = -self.__H_cp[:, :, -3, :]
+                
+                    # Boundaries in the z-direction
+                    if self.__bound[4] == 1:
+                        self.__E_cp[:, :, :, 0] = self.__E_cp[:, :, :, 1]
+                        self.__H_cp[:, :, :, 0] = self.__H_cp[:, :, :, 1]
+
+                
+                    if self.__bound[5] == 1:
+                        self.__E_cp[:, :, :, -2] = -self.__E_cp[:, :, :, -3]
+                        self.__H_cp[:, :, :, -2] = -self.__H_cp[:, :, :, -3]
+
+
+                            
         if obs:
             return obs_array
                 
@@ -1154,7 +1443,26 @@ class Continuum(object):
         start_time=time.time()
 
         if backend.upper()=="NUMPY":
+            if self.__cupy_enabled==True and (self.__conductivity or self.__conductivity_m or self.__anisotropy):
+                raise NotImplementedError("Prebuilt anisotropic or conductive medium for GPU is not available for NUMPY renderer.\nBuild for non-GPU or use 'cupy' as backend")
             obs=self.__numpy_renderer(time_steps, observers)
+            time_elapsed=time.time()-start_time
+            print(f"\nRendered Succesfully\nElapsed Time : {self.__convert_to_time(time_elapsed)} , {time_elapsed/time_steps} s/step\nRender Stream Rate : {self.__prefix_quantifier(np.prod(self.__grid_size)/time_elapsed*time_steps)}Points/s")
+            
+        elif backend.upper()=="CUPY":
+            if self.__built_for!="cupy":
+                raise ValueError("set build(gpu_flag=1) for utilizing GPU resources")
+            obs=self.__cupy_renderer(time_steps, observers)
+            time_elapsed=time.time()-start_time
+            print(f"\nRendered Succesfully\nElapsed Time : {self.__convert_to_time(time_elapsed)} , {time_elapsed/time_steps} s/step\nRender Stream Rate : {self.__prefix_quantifier(np.prod(self.__grid_size)/time_elapsed*time_steps)}Points/s")
+            cp._default_memory_pool.free_all_blocks()
+
+            print("Transfering from GPU to host memory...")
+            start_t=time.time()
+            self.__E=self.__E_cp.get()
+            self.__H=self.__H_cp.get()
+            end_t=time.time()
+            print(f"Data transferred in {end_t-start_t} secs")
         
         elif backend.upper()=="C++":
             raise ValueError("C++ renderer will be added in future releases")#self.__cpp_renderer(time_steps, observers)
@@ -1167,14 +1475,9 @@ class Continuum(object):
         
         else:
             raise ValueError(f"'{backend}' is not a recognized backend. Check for any typo. Supported background is:\n'numpy'")#raise ValueError(f"{backend} is not a recognized backend. Check for any typo. Supported backgrounds are:\n'numpy','C++','CUDA', 'FPGA")
-       
-        
-       
-       
-        
-        time_elapsed=time.time()-start_time
 
-        print(f"\nRendered Succesfully\nElapsed Time : {self.__convert_to_time(time_elapsed)} , {time_elapsed/time_steps} s/step\nRender Stream Rate : {self.__prefix_quantifier(np.prod(self.__grid_size)/time_elapsed*time_steps)}Points/s")
+        
+
         self.__rendered=True
         return obs
     
@@ -1383,4 +1686,3 @@ class Boundaries(object):
     @property
     def dimensionality(self):
         return self.__dim
-
